@@ -1,5 +1,5 @@
 # -*- mode: python; coding: utf-8 -*-
-# Copyright © 2019, 2020 Göran Weinholt
+# Copyright © 2019, 2020, 2021 Göran Weinholt
 # SPDX-License-Identifier: AGPL-3.0-or-later
 
 # This program is free software: you can redistribute it and/or modify
@@ -21,6 +21,7 @@ Sadly the Guile support in gdb was bit rotten when this was written.
 """
 
 import gdb
+import struct
 import re
 
 # Matches typetags.py and (loko arch amd64 objects)
@@ -42,10 +43,13 @@ tag_void = 0b0000000000001111
 tag_seek_mark = 0b1000000000001111 | 1<<63
 tag_box_header = 0b0100000000001111
 
-btag_bignum = 0x01000000
-btag_ratnum = 0x02000000
-btag_symbol = 0x03000000
-btag_port = 0x04000000
+btag_bignum   = 0x40000000
+btag_ratnum   = 0x43000000
+btag_symbol   = 0x01000000
+btag_port     = 0x02000000
+btag_rtd      = 0x04000000
+btag_pcompnum = 0x46000000
+btag_rcompnum = 0x49000000
 
 mask_char  = 0xff
 mask_flonum = 0xff
@@ -66,6 +70,7 @@ shift_boolean = 8
 shift_singular = 8
 shift_void = 16
 shift_boxhdr_value = 16
+shift_flonum = 32
 
 value_false = tag_boolean
 
@@ -94,6 +99,10 @@ class LokoCharPrinter(LokoPrinter):
     def to_string(self):
         return "#\\%c" % (self.val >> shift_char)
 
+class LokoFlonumPrinter(LokoPrinter):
+    def to_string(self):
+        v, = struct.unpack("=f", struct.pack("=i", (self.val >> shift_flonum)))
+        return "%f" % v
 
 class LokoVoidPrinter(LokoPrinter):
     def to_string(self):
@@ -197,12 +206,12 @@ class LokoBignumPrinter(LokoBoxPrinter):
 class LokoRecordPrinter(LokoBoxPrinter):
     def __init__(self, val, boxheader):
         super(LokoRecordPrinter, self).__init__(val, boxheader)
-        self.flags_and_len = deref(self.boxheader, 2)
-        self.name = deref(self.boxheader, 3)
-        self.parent = deref(self.boxheader, 4)
-        self.uid = deref(self.boxheader, 5)
-        self.names = deref(self.boxheader, 6)
-        self.mutable = deref(self.boxheader, 7)
+        self.name = deref(self.boxheader, 1)
+        self.record_size = deref(self.boxheader, 2)
+        self.parent = deref(self.boxheader, 3)
+        self.uid = deref(self.boxheader, 4)
+        self.names = deref(self.boxheader, 5)
+        self.mutable = deref(self.boxheader, 6)
 
     def to_string(self):
         return '#[%s]' % (self.name or self.uid)
@@ -216,8 +225,8 @@ class LokoRecordPrinter(LokoBoxPrinter):
             nonlocal i
             if rtd == value_false:
                 return
-            parent = deref(rtd, 4)
-            names = LokoVectorPrinter(deref(rtd, 6)).children()
+            parent = deref(rtd, 3)
+            names = LokoVectorPrinter(deref(rtd, 5)).children()
             yield from recurse(parent)
             for _, name in names:
                 yield "nam%d" % i, str(name)
@@ -228,30 +237,46 @@ class LokoRecordPrinter(LokoBoxPrinter):
 
 
 class LokoRtdPrinter(object):
-    def __init__(self, val):
-        flags_and_len = deref(val, 2) >> shift_fixnum
-        self.flags = flags_and_len >> 32
-        self.len = flags_and_len & 0xffffffff
-        self.name = deref(val, 3)
-        self.parent = deref(val, 4)
-        self.uid = deref(val, 5)
-        self.names = deref(val, 6)
-        self.mutable = deref(val, 7)
+    def __init__(self, val, boxheader):
+        flags = (boxheader >> shift_boxhdr_value) & mask_boxhdr_value
+        self.flags = (('O' if flags&1 else 'o') +
+                      ('S' if flags&2 else 's') +
+                      ('G' if flags&4 else 'g'))
+        self.name = deref(val, 1)
+        self.record_size = deref(val, 2) >> shift_fixnum
+        self.parent = deref(val, 3)
+        self.uid = deref(val, 4)
+        self.names = deref(val, 5)
+        self.mutable = deref(val, 6)
 
     def to_string(self):
-        return '#<rtd name=%s flags=0x%x len=%d parent=%s names=%s mutable=%d>' % (
-            self.name, self.flags, self.len, self.parent,
+        return '#<rtd name=%s flags=%s size=%d parent=%s names=%s mutable=%x>' % (
+            self.name, self.flags, self.record_size, self.parent,
             self.names, self.mutable)
 
 
+class LokoRcompnumPrinter(object):
+    def __init__(self, val, boxheader):
+        self.i = deref(val, 1)
+        self.r = deref(val, 2)
+
+    def display_hint(self):
+        return 'pcompnum'
+
+    def children(self):
+        yield "imag", self.i
+        yield "real", self.r
+
+
 class LokoProcedurePrinter(LokoPrinter):
     def __init__(self, val):
         self.val = val
         self.entry = deref(self.val, 0)
         self.info = deref(self.val, 1)
-        self.freevars = deref(self.info, 2)
-        self.name = deref(self.info, 3)
-        self.source = deref(self.info, 4)
+
+        self.name = deref(self.info, 0+1)
+        self.freevars = deref(self.info, 1+1)
+        self.source = deref(self.info, 2+1)
 
     def to_string(self):
         if self.freevars != 0:
@@ -260,12 +285,14 @@ class LokoProcedurePrinter(LokoPrinter):
             return "#<procedure>"
 
     def children(self):
-        yield 'name', self.name
-        # FIXME: This should be returned as a linker symbol
-        yield 'entry', hex(self.entry)
-        yield 'source', self.source
-        yield 'num-free', self.freevars
-        # FIXME: get all free variables
+        yield 'entry', self.entry.cast(gdb.lookup_type("void").pointer())
+        yield 'info', self.info
+
+        # TODO: Enable this once top-level values are not part of
+        # closures. For now it's just too spammy.
+
+        # for i in range(self.freevars):
+        #     yield "free%i" % i, deref(self.val, 2+i)
 
 
 class LokoImmsymPrinter(LokoPrinter):
@@ -329,15 +356,9 @@ def loko_lookup_function(val):
 
         if tag3 == tag_box:
             boxheader = deref(val, 0)
-            if int(boxheader) & 0b1111 == tag_immsym:
-                # FIXME: Obsolete
-                oldtype = LokoImmsymPrinter(boxheader).to_string()
-                if oldtype == 'rtd':
-                    return LokoRtdPrinter(val)
             if int(boxheader) & 0b111 == tag_box:
                 boxheader2 = deref(boxheader, 0)
-                oldtype2 = LokoImmsymPrinter(boxheader2).to_string()
-                if oldtype2 == 'rtd':
+                if int(boxheader2) & mask_boxhdr_type == btag_rtd:
                     return LokoRecordPrinter(val, boxheader)
             if int(boxheader) & mask_box_header == tag_box_header:
                 if int(boxheader) & mask_boxhdr_type == btag_symbol:
@@ -349,6 +370,10 @@ def loko_lookup_function(val):
                     return LokoPortPrinter(val, boxheader)
                 elif int(boxheader) & mask_boxhdr_type == btag_bignum:
                     return LokoBignumPrinter(val, boxheader)
+                elif int(boxheader) & mask_boxhdr_type == btag_rtd:
+                    return LokoRtdPrinter(val, boxheader)
+                elif int(boxheader) & mask_boxhdr_type == btag_rcompnum:
+                    return LokoRcompnumPrinter(val, boxheader)
 
             return LokoBoxPrinter(val, boxheader)
 
@@ -373,6 +398,8 @@ def loko_lookup_function(val):
                 return LokoSingularPrinter('#t')
         if tag8 == tag_char:
             return LokoCharPrinter(val)
+        if tag8 == tag_flonum:
+            return LokoFlonumPrinter(val)
 
         tag16 = int(val) & 0xffff
         if tag16 == tag_void:
@@ -429,7 +456,7 @@ class LokoUnwinder(Unwinder):
 
 gdb.pretty_printers.append(loko_lookup_function)
 
-# For some reason gdb.lookup_global_symbol does not.
+# For some reason gdb.lookup_global_symbol does not work.
 table = gdb.parse_and_eval("&bootstrap_unwind_table")
 if table is None:
     print("Failed to get the address of bootstrap_unwind_table")
